
/* standard C headers {{{ */
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <assert.h>
/* }}} */

/* standard POSIX headers {{{ */
#define __USE_BSD       /* for gethostname() in unistd.h */
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>
#include <libgen.h>     /* basename() */
/* }}} */

/* local definitions {{{ */
#include "xterm-title.h"

/* You should _NOT_ change these values. */
#define XTERM_TITLE_DEFAULT_FORMAT "%U: %r (%p)"
#define XTERM_TITLE_SIZE 4096
#define XTERM_TITLE_ENV_VAR "XTERM_TITLE"
#define XTERM_TITLE_PATH_SEP '/'

/* Buffer size for maximum path. */
#ifdef MAXPATHLEN
#define XTERM_TITLE_PATH_SIZE (MAXPATHLEN + 1)
#elif defined (PATH_MAX)
#define XTERM_TITLE_PATH_SIZE (PATH_MAX + 1)
#else
#define XTERM_TITLE_PATH_SIZE 8192
#endif

/* Buffer size for maximum hostname. */
#ifdef MAXHOSTNAMELEN
#define XTERM_TITLE_HOSTNAME_SIZE (MAXHOSTNAMELEN + 1)
#elif defined (HOST_NAME_MAX)
#define XTERM_TITLE_HOSTNAME_SIZE (HOST_NAME_MAX + 1)
#else
#define XTERM_TITLE_HOSTNAME_SIZE 8192
#endif
/* }}} */

/* declarations of static functions {{{ */
static char * get_working_dir (void);
static char * get_hostname (void);
static const char * get_homedir (void);
static const char * get_username (void);
static char * strdup_upper (const char *src);
static char * strdup_lower (const char *src);
static char * strdup_case (const char *src, int (*func)(int c));
static char * homedir_to_tilde (const char *path);
static int write_title (FILE *tty, const char *appname, const char *resource);
/* }}} */

int xterm_title_set (const char *appname, const char *resource, /* {{{ */
        int force)
{
    FILE *tty;
    char *env_term;
    
    if (force || getenv(XTERM_TITLE_ENV_VAR) || ((env_term = getenv("TERM"))
            && (!strcmp(env_term, "xterm") 
                || !strcmp(env_term, "rxvt")))) {
        /* Terminal is xterm compatible. */
        /* Open the terminal. */
        if ((tty = fopen("/dev/tty", "w+"))) {
            int status = write_title(tty, appname, resource);
            fclose(tty);
            return (status);
        }
        else {
            return (0); /* Cannot open "/dev/tty". */
        }
    }
    else {
        return (0); /* Terminal is _NOT_ xterm compatible. */
    }
}
/* }}} */

static int write_title (FILE *tty, const char *appname, /* {{{ */
        const char *resource)
{
    const char *format;
    const char *env_title;
    const char *s;
    char buf[XTERM_TITLE_SIZE];
    int esc, i;
    
    if (tty == NULL || appname == NULL)
        return (0);

    if ((env_title = getenv("XTERM_TITLE")))
        format = env_title;
    else
        format = XTERM_TITLE_DEFAULT_FORMAT; /* from util.h */

#define ADD_BUF(str) \
    if (str) { \
        int len = strlen(str); \
        int rest = sizeof(buf) - i - 1; \
        int cpy =  len > rest ? rest : len; \
        memcpy(&buf[i], str, cpy);  i += cpy; }
    
    esc = i = 0;
    for (s = format; *s != '\0' && i < sizeof(buf); s++) {
        char ch = *s;
        if (esc) {
            switch (ch) {
                /* escaped '%' character */
                case '%':
                    buf[i++] = '%';
                    break;

                /* application name */
                case 'a':
                    ADD_BUF(appname);
                    break;
                
                /* lower case application name */
                case 'L':
                    {
                        char *t = strdup_lower(appname);
                        ADD_BUF(t);
                        free(t);
                    }
                    break;
                
                /* upper case application name */
                case 'U':
                    {
                        char *t = strdup_upper(appname);
                        ADD_BUF(t);
                        free(t);
                    }
                    break;
                
                /* active resource */
                case 'r':
                    ADD_BUF(resource);
                    break;
                
                /* process ID */
                case 'p':
                    {
                        pid_t pid = getpid();
                        char ptmp[256];
                        sprintf(ptmp, "%u", pid); /* safe */
                        ADD_BUF(ptmp);
                    }
                    break;
                
                /* username */
                case 'u':
                    {
                        const char *username;
                        if ((username = get_username())) {
                            ADD_BUF(username);
                        }
                        else
                            return (0);
                    }
                    break;
               
                /* current working directory - full path
                 * not in the specification */
                case 'w':
                    {
                        char *wd = get_working_dir();
                        if (wd == NULL)
                            return (0);
                        else {
                            char *wd_tilde = homedir_to_tilde(wd);
                            ADD_BUF(wd_tilde);
                        }
                    }
                    break;
 
                /* current working directory - basename
                 * not in the specification */
                case 'W':
                    {
                        char *wd = get_working_dir();
                        if (wd == NULL)
                            return (0);
                        else
                            ADD_BUF(basename(wd));
                    }
                    break;
 
                /* hostname up to the first dot */
                case 'h':
                    {
                        char *hn = get_hostname();
                        if (hn == NULL)
                            return (0);
                        else {
                            char *dot = strchr(hn, '.');
                            if (dot) *dot = '\0';
                            ADD_BUF(hn);
                        }
                    }
                    break;
 
                /* hostname */
                case 'H':
                    {
                        char *hn = get_hostname();
                        if (hn == NULL)
                            return (0);
                        else
                            ADD_BUF(hn);
                    }
                    break;
                 
                /* unknown escape code, ignore it */
                default:
                    ;
            }
            esc = 0;
        }
        else {
            if (ch == '%')
                esc = 1;
            else
                buf[i++] = ch;
        }
    }
    buf[i] = '\0';
    
    fprintf(tty, "\033]0;%s\a", buf);
    fflush(tty);
    return (ferror(tty) ? 0 : 1);
}
/* }}} */

static char * get_working_dir (void) /* {{{ */
{
    static char wd[XTERM_TITLE_PATH_SIZE];
    char *r;
    /* We don't use getcwd(3) on SUNOS, because, it does 
     * a popen("pwd") and, if that wasn't bad enough, 
     * hangs in doing so. (wisdom from glib)
     */ 
#if defined (sun) && !defined (__SVR4)
    r = getwd(wd);
#else
    r = getcwd(wd, XTERM_TITLE_PATH_SIZE);
#endif  /* sun */
    
    if (r == NULL)
        return (NULL);
    else {
        wd[sizeof(wd)-1] = '\0'; /* be safe */
        return (wd);
    }
}
/* }}} */

static char * get_hostname (void) /* {{{ */
{
    static char hn[XTERM_TITLE_HOSTNAME_SIZE];
    char *env_hostname;
    
    if ((env_hostname = getenv("HOSTNAME"))) {
        strncpy(hn, env_hostname, sizeof(hn));
        hn[sizeof(hn)-1] = '\0';
        return (hn);
    }
    else {
        if (gethostname(hn, XTERM_TITLE_HOSTNAME_SIZE) == -1)
            return (NULL);
        else {
            hn[sizeof(hn)-1] = '\0'; /* be safe */
            return (hn);
        }
    }
}
/* }}} */

static const char * get_username (void) /* {{{ */
{
    char *env_user;
    struct passwd *pw;

    if ((env_user = getenv("USER")) || (env_user = getenv("USERNAME"))) {
        return (env_user);
    }
    else {
        if ((pw = getpwuid(getuid()))) {
            return (pw->pw_name);
        }
        else {
            return (NULL);
        }
    }
}
/* }}} */

static const char * get_homedir (void) /* {{{ */
{
    struct passwd *pw;
    char *env_home;
    
    if ((env_home = getenv("HOME")))
        return (env_home);
    else {
        if ((pw = getpwuid(getuid())) == NULL)
            return (NULL);
        else
            return (pw->pw_dir);
    }
}
/* }}} */

static char * homedir_to_tilde (const char *path) /* {{{ */
{
    static char npath[XTERM_TITLE_PATH_SIZE];
    int len;
    const char *homedir;
    
    if (path == NULL)
        return (NULL);

    if ((homedir = get_homedir()) == NULL)
        return ((char *)path);
    
    if (! strncmp(path, homedir, len = strlen(homedir))
            && (path[len] == XTERM_TITLE_PATH_SEP || path[len] == '\0')) {
    	npath[0] = '~';
	    strncpy(&npath[1], &path[len], sizeof(npath));
        npath[sizeof(npath)-1] = '\0';
    	return (npath);
    }
    else {
        return ((char *)path);
    }
}
/* }}} */

static char * strdup_case (const char *src, int (*func)(int c)) /* {{{ */
{
    char *dst;
    int i;
    
    assert(src);
    if ((dst = malloc(strlen(src) + 1)) == NULL)
        return (NULL);
    
    for (i = 0; src[i] != '\0'; i++) {
        dst[i] = func(src[i]);
    }
    dst[i] = '\0';
    
    return (dst);
}
/* }}} */

static char * strdup_upper (const char *src) /* {{{ */
{
    return (strdup_case(src, toupper));
}
/* }}} */

static char * strdup_lower (const char *src) /* {{{ */
{
    return (strdup_case(src, tolower));
}
/* }}} */

/* vim: set ts=4 sw=4 et: */
